"""Rule to detect presence or absence."""

import logging
import threading
import typing

import HABApp.openhab.definitions
import HABApp.openhab.events
import HABApp.openhab.items
import HABApp.util

import habapp_rules.core.helper
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.system.config.presence
from habapp_rules.system import PresenceState

LOGGER = logging.getLogger(__name__)


class Presence(habapp_rules.core.state_machine_rule.StateMachineRule):
    """Rule class to manage the presence of a home.

    Hint: If you have some kind of guest-mode, use a guest-available switch as a phone to enable a persistent presence, also if all phones are not at home

    Example OpenHAB configuration:
    # KNX-things:
    Thing device T00_99_OpenHab_Presence "KNX OpenHAB Presence"{
        Type switch-control        : presence       "Presence"      [ ga="0/2/11+0/2/10"]
        Type switch-control        : leaving        "Leaving"       [ ga="0/2/21+0/2/20"]
    }

    # Items:
    Switch    I01_00_Presence    "Presence"         <presence>    (G00_00_rrd4j)    ["Status", "Presence"]    {channel="knx:device:bridge:T00_99_OpenHab_Presence:presence"}
    Switch    I01_00_Leaving     "Leaving"          <leaving>                                                 {channel="knx:device:bridge:T00_99_OpenHab_Presence:leaving"}

    # Config:
    config = habapp_rules.system.config.presence.PresenceConfig(
            items=habapp_rules.system.config.presence.PresenceItems(
                    presence="I01_00_Presence",
                    leaving="I01_00_Leaving"
            )
    )

    # Rule init:
    habapp_rules.system.presence.Presence(config)
    """

    states: typing.ClassVar = [
        {"name": "presence"},
        {"name": "leaving", "timeout": 5 * 60, "on_timeout": "absence_detected"},  # leaving takes 5 minutes
        {"name": "absence", "timeout": 1.5 * 24 * 3600, "on_timeout": "long_absence_detected"},  # switch to long absence after 1.5 days
        {"name": "long_absence"},
    ]

    trans: typing.ClassVar = [
        {"trigger": "presence_detected", "source": ["absence", "long_absence"], "dest": "presence"},
        {"trigger": "leaving_detected", "source": ["presence", "absence", "long_absence"], "dest": "leaving"},
        {"trigger": "abort_leaving", "source": "leaving", "dest": "presence"},
        {"trigger": "absence_detected", "source": ["presence", "leaving"], "dest": "absence"},
        {"trigger": "long_absence_detected", "source": "absence", "dest": "long_absence"},
    ]

    def __init__(self, config: habapp_rules.system.config.presence.PresenceConfig) -> None:
        """Init of Presence object.

        Args:
            config: config for presence detection
        """
        self._config = config
        habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, config.items.state)
        self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, config.items.presence.name)

        # init state machine
        self.state_machine = habapp_rules.core.state_machine_rule.StateMachineWithTimeout(model=self, states=self.states, transitions=self.trans, ignore_invalid_triggers=True, after_state_change="_update_openhab_state")
        self._set_initial_state()

        # add callbacks
        self._config.items.leaving.listen_event(self._cb_leaving, HABApp.openhab.events.ItemStateChangedEventFilter())
        self._config.items.presence.listen_event(self._cb_presence, HABApp.openhab.events.ItemStateChangedEventFilter())
        for door_item in self._config.items.outdoor_doors:
            door_item.listen_event(self._cb_outside_door, HABApp.core.events.ValueChangeEventFilter())
        for phone_item in self._config.items.phones:
            phone_item.listen_event(self._cb_phone, HABApp.core.events.ValueChangeEventFilter())

        self.__phone_absence_timer: threading.Timer | None = None
        self._instance_logger.debug(super().get_initial_log_message())

    def _get_initial_state(self, default_value: str = PresenceState.PRESENCE.value) -> str:
        """Get initial state of state machine.

        Args:
            default_value: default / initial state

        Returns:
            return correct state if it could be detected, if not return default value
        """
        phone_items = [phone for phone in self._config.items.phones if phone.value is not None]  # phones with valid state (not None)
        if phone_items:
            if any(item.value == "ON" for item in phone_items):
                return PresenceState.PRESENCE.value

            if self._config.items.presence.value == "ON":
                return PresenceState.LEAVING.value
            return PresenceState.LONG_ABSENCE.value if self._item_state.value == PresenceState.LONG_ABSENCE.value else PresenceState.ABSENCE.value

        if self._config.items.leaving.value == "ON":
            return PresenceState.LEAVING.value

        if self._config.items.presence.value == "ON":
            return PresenceState.PRESENCE.value

        if self._config.items.presence.value == "OFF":
            return PresenceState.LONG_ABSENCE.value if self._item_state.value == PresenceState.LONG_ABSENCE.value else PresenceState.ABSENCE.value

        return default_value

    def _update_openhab_state(self) -> None:
        """Extend _update_openhab state of base class to also update other OpenHAB items."""
        super()._update_openhab_state()
        self._instance_logger.info(f"Presence state changed to {self.state}")

        # update presence item
        target_value = "ON" if self.state in {PresenceState.PRESENCE.value, PresenceState.LEAVING.value} else "OFF"
        habapp_rules.core.helper.send_if_different(self._config.items.presence, target_value)
        habapp_rules.core.helper.send_if_different(self._config.items.leaving, "ON" if self.state == PresenceState.LEAVING.value else "OFF")

    def _cb_outside_door(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is called if any outside door changed state.

        Args:
            event: state change event of door item
        """
        if event.value == "OPEN" and self.state not in {PresenceState.PRESENCE.value, PresenceState.LEAVING.value}:
            self._instance_logger.debug(f"Presence detected by door ({event.name})")
            self.presence_detected()

    def _cb_leaving(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is called if leaving item changed state.

        Args:
            event: Item state change event of leaving item
        """
        if event.value == "ON" and self.state in {PresenceState.PRESENCE.value, PresenceState.ABSENCE.value, PresenceState.LONG_ABSENCE.value}:
            self._instance_logger.debug("Start leaving through leaving switch")
            self.leaving_detected()
        if event.value == "OFF" and self.state == PresenceState.LEAVING.value:
            self._instance_logger.debug("Abort leaving through leaving switch")
            self.abort_leaving()

    def _cb_presence(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is called if presence item changed state.

        Args:
            event: Item state change event of presence item
        """
        if event.value == "ON" and self.state in {PresenceState.ABSENCE.value, PresenceState.LONG_ABSENCE.value}:
            self._instance_logger.debug("Presence was set manually by presence switch")
            self.presence_detected()
        elif event.value == "OFF" and self.state in {PresenceState.PRESENCE.value, PresenceState.LEAVING.value}:
            self._instance_logger.debug("Absence was set manually by presence switch")
            self.absence_detected()

    def _cb_phone(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is called if a phone state changed.

        Args:
            event: Item state change event of phone item
        """
        active_phones = len([phone for phone in self._config.items.phones if phone.value == "ON"])
        if active_phones == 1 and event.value == "ON":
            # first phone switched to ON
            if self.__phone_absence_timer:
                self.__phone_absence_timer.cancel()
                self.__phone_absence_timer = None

            if self.state == PresenceState.LEAVING.value:
                self._instance_logger.debug("Leaving was aborted through first phone which came online")
                self.abort_leaving()

            if self.state in {PresenceState.ABSENCE.value, PresenceState.LONG_ABSENCE.value}:
                self._instance_logger.debug("Presence was set through first phone joined network")
                self.presence_detected()

        elif active_phones == 0 and event.value == "OFF":
            # last phone switched to OFF
            self.__phone_absence_timer = threading.Timer(20 * 60, self.__set_leaving_through_phone)
            self.__phone_absence_timer.start()

    def __set_leaving_through_phone(self) -> None:
        """Set leaving detected if timeout expired."""
        if self.state == PresenceState.PRESENCE.value:
            self._instance_logger.debug("Leaving was set, because last phone left some time ago.")
            self.leaving_detected()
        self.__phone_absence_timer = None

    def on_rule_removed(self) -> None:
        """Stop all threads, started by this rule."""
        habapp_rules.core.state_machine_rule.StateMachineRule.on_rule_removed(self)

        # stop phone absence timer
        if self.__phone_absence_timer:
            self.__phone_absence_timer.cancel()
            self.__phone_absence_timer = None



"""Tests for Watchdog Rule."""

import unittest.mock

import HABApp.openhab.items

import habapp_rules.system.item_watchdog
import tests.helper.oh_item
import tests.helper.test_case_base
from habapp_rules.system.config.item_watchdog import WatchdogConfig, WatchdogItems, WatchdogParameter


class TestWatchdog(tests.helper.test_case_base.TestCaseBase):
    """Tests for Watchdog Rule."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Number", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Switch", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Number_Warning", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Switch_Warning", None)

        self._watchdog_number = habapp_rules.system.item_watchdog.ItemWatchdog(WatchdogConfig(items=WatchdogItems(observed="Unittest_Number", warning="Unittest_Number_Warning")))

        self._watchdog_switch = habapp_rules.system.item_watchdog.ItemWatchdog(WatchdogConfig(items=WatchdogItems(observed="Unittest_Switch", warning="Unittest_Switch_Warning"), parameter=WatchdogParameter(timeout=10)))

    def test_cb_observed_state_updated(self) -> None:
        """Callback which is called if the observed item was updated."""
        with unittest.mock.patch.object(self._watchdog_number, "_countdown") as number_countdown_mock, unittest.mock.patch.object(self._watchdog_switch, "_countdown") as switch_countdown_mock:
            tests.helper.oh_item.item_state_event("Unittest_Number", 42)
            number_countdown_mock.reset.assert_called_once()
            switch_countdown_mock.reset.assert_not_called()
            tests.helper.oh_item.assert_value("Unittest_Number_Warning", "OFF")
            tests.helper.oh_item.assert_value("Unittest_Switch_Warning", None)

            tests.helper.oh_item.item_state_event("Unittest_Switch", "OFF")
            number_countdown_mock.reset.assert_called_once()
            switch_countdown_mock.reset.assert_called_once()
            tests.helper.oh_item.assert_value("Unittest_Number_Warning", "OFF")
            tests.helper.oh_item.assert_value("Unittest_Switch_Warning", "OFF")



"""Tests for monthly energy report."""

import collections
import datetime
import unittest
import unittest.mock

import HABApp.openhab.definitions.helpers.persistence_data
import HABApp.openhab.items
import multi_notifier.connectors.connector_mail

import habapp_rules
import habapp_rules.core.exceptions
import habapp_rules.energy.config.monthly_report
import habapp_rules.energy.monthly_report
import tests.helper.oh_item
import tests.helper.test_case_base


class TestFunctions(unittest.TestCase):
    """Test all global functions."""

    def test_get_last_month_name(self) -> None:
        """Test _get_last_month_name."""
        TestCase = collections.namedtuple("TestCase", ["month_number", "expected_name"])

        test_cases = [
            TestCase(1, "Dezember"),
            TestCase(2, "Januar"),
            TestCase(3, "Februar"),
            TestCase(4, "März"),
            TestCase(5, "April"),
            TestCase(6, "Mai"),
            TestCase(7, "Juni"),
            TestCase(8, "Juli"),
            TestCase(9, "August"),
            TestCase(10, "September"),
            TestCase(11, "Oktober"),
            TestCase(12, "November"),
        ]

        today = datetime.datetime.today()
        with unittest.mock.patch("datetime.datetime") as datetime_mock:
            for test_case in test_cases:
                with self.subTest(test_case=test_case):
                    datetime_mock.now.return_value = today.replace(month=test_case.month_number, day=1)
                    self.assertEqual(test_case.expected_name, habapp_rules.energy.monthly_report._get_previous_month_name())


class TestMonthlyReport(tests.helper.test_case_base.TestCaseBase):
    """Test MonthlyReport rule."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Energy_Sum", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Energy_1", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Energy_2", None)

        self._energy_1 = habapp_rules.energy.config.monthly_report.EnergyShare("Energy_1", "Energy 1")
        self._energy_2 = habapp_rules.energy.config.monthly_report.EnergyShare("Energy_2", "Energy 2")
        self._mail_config = multi_notifier.connectors.connector_mail.MailConfig(user="User", password="Password", smtp_host="smtp.test.de", smtp_port=587)  # noqa: S106

        config = habapp_rules.energy.config.monthly_report.MonthlyReportConfig(
            items=habapp_rules.energy.config.monthly_report.MonthlyReportItems(energy_sum="Energy_Sum"),
            parameter=habapp_rules.energy.config.monthly_report.MonthlyReportParameter(known_energy_shares=[self._energy_1, self._energy_2], config_mail=self._mail_config, recipients=["test@test.de"]),
        )

        self._rule = habapp_rules.energy.monthly_report.MonthlyReport(config)

    def test_init(self) -> None:
        """Test init."""
        TestCase = collections.namedtuple("TestCase", ["sum_in_group", "item_1_in_group", "item_2_in_group", "raises_exception"])

        test_cases = [
            TestCase(True, True, True, False),
            TestCase(True, True, False, True),
            TestCase(True, False, True, True),
            TestCase(True, False, False, True),
            TestCase(False, True, True, True),
            TestCase(False, True, False, True),
            TestCase(False, False, True, True),
            TestCase(False, False, False, True),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                self._energy_1.energy_item.groups = {"PersistenceGroup"} if test_case.item_1_in_group else set()
                self._energy_2.energy_item.groups = {"PersistenceGroup"} if test_case.item_2_in_group else set()
                self._rule._config.items.energy_sum.groups = {"PersistenceGroup"} if test_case.sum_in_group else set()

                config = habapp_rules.energy.config.monthly_report.MonthlyReportConfig(
                    items=habapp_rules.energy.config.monthly_report.MonthlyReportItems(energy_sum="Energy_Sum"),
                    parameter=habapp_rules.energy.config.monthly_report.MonthlyReportParameter(
                        known_energy_shares=[self._energy_1, self._energy_2], config_mail=self._mail_config, recipients=["test@test.de"], persistence_group_name="PersistenceGroup"
                    ),
                )

                if test_case.raises_exception:
                    with self.assertRaises(habapp_rules.core.exceptions.HabAppRulesConfigurationError):
                        habapp_rules.energy.monthly_report.MonthlyReport(config)
                else:
                    habapp_rules.energy.monthly_report.MonthlyReport(config)

    def test_init_with_debug_mode(self) -> None:
        """Test init with debug mode."""
        config = habapp_rules.energy.config.monthly_report.MonthlyReportConfig(
            items=habapp_rules.energy.config.monthly_report.MonthlyReportItems(energy_sum="Energy_Sum"),
            parameter=habapp_rules.energy.config.monthly_report.MonthlyReportParameter(known_energy_shares=[self._energy_1, self._energy_2], config_mail=self._mail_config, recipients=["test@test.de"], debug=True),
        )

        self._rule = habapp_rules.energy.monthly_report.MonthlyReport(config)

    def test_create_html(self) -> None:
        """Test create_html."""
        self._rule._config.items.energy_sum.value = 20_123.5489135

        template_mock = unittest.mock.MagicMock()
        with unittest.mock.patch("jinja2.Template", return_value=template_mock), unittest.mock.patch("habapp_rules.energy.monthly_report._get_previous_month_name", return_value="MonthName"):
            self._rule._create_html(10_042.123456)

        template_mock.render.assert_called_once_with(month="MonthName", energy_now="20123.5", energy_last_month="10042.1", habapp_version=habapp_rules.__version__, chart="{{ chart }}")

    def test_cb_send_energy(self) -> None:
        """Test cb_send_energy."""
        self._rule._config.items.energy_sum.value = 1000
        self._energy_1.energy_item.value = 100
        self._energy_2.energy_item.value = 50

        with (
            unittest.mock.patch("habapp_rules.energy.helper.get_historic_value", side_effect=[800, 90, 45]),
            unittest.mock.patch("habapp_rules.energy.donut_chart.create_chart", return_value="html text result") as create_chart_mock,
            unittest.mock.patch.object(self._rule, "_create_html") as create_html_mock,
            unittest.mock.patch("habapp_rules.energy.monthly_report._get_previous_month_name", return_value="MonthName"),
            unittest.mock.patch.object(self._rule, "_mail") as mail_mock,
        ):
            self._rule._cb_send_energy()

        create_chart_mock.assert_called_once_with(["Energy 1", "Energy 2", "Rest"], [10, 5, 185], unittest.mock.ANY)
        create_html_mock.assert_called_once_with(200)
        mail_mock.send_message("test@test.de", "html text result", "Stromverbrauch MonthName", images={"chart": unittest.mock.ANY})



"""Test for timeout_list."""

import dataclasses
import time
from collections.abc import Iterator
from typing import Any


@dataclasses.dataclass
class ValueWithTimeout:
    """Define item for TimeoutList."""

    value: object
    timeout: float
    add_timestamp: float

    def __lt__(self, other: "ValueWithTimeout") -> bool:
        """Compare two ValueWithTimeout objects.

        Args:
            other: other object for comparison

        Returns:
            True if self.value is less than other.value
        """
        if not isinstance(other, ValueWithTimeout):
            return NotImplemented
        return self.value < other.value

    def __eq__(self, other: object) -> bool:
        """Check if equal.

        Args:
            other: other item

        Returns:
            True if equal
        """
        if not isinstance(other, ValueWithTimeout):
            return NotImplemented
        return self.value == other.value

    def __hash__(self) -> int:
        """Get hash of the ValueWithTimeout object.

        Returns:
            hash value
        """
        return hash((self.value, self.timeout, self.add_timestamp))


class TimeoutList:
    """List like class, where every item has a timeout, which will remove it from the list."""

    def __init__(self) -> None:
        """Init class."""
        self.__items: list[ValueWithTimeout] = []

    def __repr__(self) -> str:
        """Get representation of all list items (without timeout).

        Returns:
            all list elements which are currently in the list
        """
        self.__remove_old_items()
        return str([itm.value for itm in self.__items])

    def __bool__(self) -> bool:
        """Check if list has items.

        Returns:
            true if items in list
        """
        self.__remove_old_items()
        return bool(self.__items)

    def __contains__(self, item: object) -> bool:
        """Check if an item is in the list.

        Args:
            item: item which should be checked

        Returns:
            true if item is in the list
        """
        self.__remove_old_items()
        return item in [itm.value for itm in self.__items]

    def __getitem__(self, index: int) -> object:
        """Get item from list by index.

        Args:
            index: index of item position in list.

        Returns:
            item from list
        """
        self.__remove_old_items()
        return self.__items[index].value

    def __eq__(self, other: object) -> bool:
        """Check if equal.

        Args:
            other: other item

        Returns:
            true if equal
        """
        if isinstance(other, TimeoutList):
            return repr(self) == repr(other)

        if isinstance(other, list):
            self.__remove_old_items()
            return [itm.value for itm in self.__items] == other

        return False

    def __hash__(self) -> int:
        """Get hash of the TimeoutList object.

        Returns:
            hash value
        """
        return hash(tuple(itm.value for itm in self.__items))

    def __iter__(self) -> Iterator[Any]:
        """Get an iterator for the items in the list.

        The iterator yields the values of the items in the list.

        Returns:
            iterator for the items in the list
        """
        self.__remove_old_items()
        return (itm.value for itm in self.__items)

    def __remove_old_items(self) -> None:
        """Remove items from list, which are timed-out."""
        current_time = time.time()
        self.__items = [itm for itm in self.__items if current_time - itm.add_timestamp < itm.timeout]

    def append(self, item: object, timeout: float) -> None:
        """Add item to list.

        Args:
            item: item which should be added to the list
            timeout: timeout, after which the item is not valid anymore
        """
        self.__items.append(ValueWithTimeout(item, timeout, time.time()))

    def remove(self, item: object) -> None:
        """Remove item from list. If there are duplicates. The first element will be removed.

        Args:
            item: item which should be deleted

        Raises:
            ValueError: if item not in list
        """
        item_to_remove = next((itm for itm in self.__items if itm.value == item), None)

        if not item_to_remove:
            msg = f"{self.__class__.__name__}.remove(x): x not in list"
            raise ValueError(msg)

        self.__items.remove(item_to_remove)

    def pop(self, element_index: int) -> object:
        """Pop item from list.

        Args:
            element_index: list index of element which should be deleted

        Returns:
            item which was removed

        Raises:
            IndexError: if index is out of range
            TypeError: if index can not be interpreted as integer
        """
        return self.__items.pop(element_index).value



"""Config models for presence rules."""

import HABApp
import pydantic

import habapp_rules.core.pydantic_base


class PresenceItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for presence detection."""

    presence: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="presence item")
    leaving: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="leaving item")
    outdoor_doors: list[HABApp.openhab.items.ContactItem] = pydantic.Field([], description="list of door contacts which are used to detect presence if outside door was opened")
    phones: list[HABApp.openhab.items.SwitchItem] = pydantic.Field([], description="list of phone items which are used to detect presence and leaving depending on present phones")
    state: HABApp.openhab.items.StringItem = pydantic.Field(..., description="state item")


class PresenceConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for presence detection."""

    items: PresenceItems = pydantic.Field(..., description="items for presence detection")
    parameter: None = None



"""Energy save switch rules."""

import logging
import typing

import HABApp

import habapp_rules.actors.config.energy_save_switch
import habapp_rules.actors.state_observer
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
from habapp_rules.system import PresenceState, SleepState

LOGGER = logging.getLogger(__name__)


class EnergySaveSwitch(habapp_rules.core.state_machine_rule.StateMachineRule):
    """Rules class to manage energy save switches.

    # Items:
    Switch    Switch              "Switch"        {channel="..."}
    Switch    Switch_State        "State"

    # Config:
    config = habapp_rules.actors.config.energy_save_switch.EnergySaveSwitchConfig(
            items=EnergySaveSwitchItems(
                    switch="Switch",
                    state="Switch_State"
            )
    ))

    # Rule init:
    habapp_rules.actors.energy_save_switch.EnergySaveSwitch(config)
    """

    states: typing.ClassVar = [
        {"name": "Manual"},
        {"name": "Hand", "timeout": 0, "on_timeout": ["_auto_hand_timeout"]},
        {
            "name": "Auto",
            "initial": "Init",
            "children": [
                {"name": "Init"},
                {"name": "On"},
                {"name": "Off"},
                {"name": "WaitCurrent"},
                {"name": "WaitCurrentExtended", "timeout": 0, "on_timeout": ["extended_wait_timeout"]},
            ],
        },
    ]

    trans: typing.ClassVar = [
        # manual
        {"trigger": "manual_on", "source": ["Auto", "Hand"], "dest": "Manual"},
        {"trigger": "manual_off", "source": "Manual", "dest": "Auto"},
        # hand
        {"trigger": "hand_detected", "source": "Auto", "dest": "Hand"},
        {"trigger": "_auto_hand_timeout", "source": "Hand", "dest": "Auto"},
        # sleeping presence conditions
        {"trigger": "on_conditions_met", "source": ["Auto_Off", "Auto_WaitCurrent", "Auto_WaitCurrentExtended"], "dest": "Auto_On"},
        {"trigger": "off_conditions_met", "source": "Auto_On", "dest": "Auto_Off", "unless": "_current_above_threshold"},
        {"trigger": "off_conditions_met", "source": "Auto_On", "dest": "Auto_WaitCurrent", "conditions": "_current_above_threshold"},
        # switch off
        {"trigger": "current_below_threshold", "source": "Auto_WaitCurrent", "dest": "Auto_WaitCurrentExtended"},
        {"trigger": "current_above_threshold", "source": "Auto_WaitCurrentExtended", "dest": "Auto_WaitCurrent"},
        {"trigger": "extended_wait_timeout", "source": "Auto_WaitCurrentExtended", "dest": "Auto_Off"},
        {"trigger": "max_on_countdown", "source": ["Auto_On", "Auto_WaitCurrent", "Hand"], "dest": "Auto_Off"},
    ]

    def __init__(self, config: habapp_rules.actors.config.energy_save_switch.EnergySaveSwitchConfig) -> None:
        """Init of energy save switch.

        Args:
            config: energy save switch config
        """
        self._config = config
        self._switch_observer = habapp_rules.actors.state_observer.StateObserverSwitch(config.items.switch.name, self._cb_hand, self._cb_hand)

        habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, self._config.items.state)
        self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, self._config.items.switch.name)

        # init state machine
        self._previous_state = None
        self.state_machine = habapp_rules.core.state_machine_rule.HierarchicalStateMachineWithTimeout(model=self, states=self.states, transitions=self.trans, ignore_invalid_triggers=True, after_state_change="_update_openhab_state")

        self._max_on_countdown = self.run.countdown(self._config.parameter.max_on_time, self._cb_max_on_countdown) if self._config.parameter.max_on_time is not None else None
        self._switch_off_after_external_req = False
        self._set_timeouts()
        self._set_state(self._get_initial_state())

        # callbacks
        self._config.items.switch.listen_event(self._cb_switch, HABApp.openhab.events.ItemStateChangedEventFilter())

        if self._config.items.manual is not None:
            self._config.items.manual.listen_event(self._cb_manual, HABApp.openhab.events.ItemStateChangedEventFilter())
        if self._config.items.presence_state is not None:
            self._config.items.presence_state.listen_event(self._cb_presence_state, HABApp.openhab.events.ItemStateChangedEventFilter())
        if self._config.items.sleeping_state is not None:
            self._config.items.sleeping_state.listen_event(self._cb_sleeping_state, HABApp.openhab.events.ItemStateChangedEventFilter())
        if self._config.items.external_request is not None:
            self._config.items.external_request.listen_event(self._cb_external_request, HABApp.openhab.events.ItemStateChangedEventFilter())
        if self._config.items.current is not None:
            self._config.items.current.listen_event(self._cb_current_changed, HABApp.openhab.events.ItemStateChangedEventFilter())

        LOGGER.info(self.get_initial_log_message())

    def _set_timeouts(self) -> None:
        """Set timeouts."""
        self.state_machine.states["Hand"].timeout = self._config.parameter.hand_timeout or 0
        self.state_machine.states["Auto"].states["WaitCurrentExtended"].timeout = self._config.parameter.extended_wait_for_current_time

    def _get_initial_state(self, default_value: str = "") -> str:  # noqa: ARG002
        """Get initial state of state machine.

        Args:
          default_value: default / initial state

        Returns:
          if OpenHAB item has a state it will return it, otherwise return the given default value

        """
        if self._config.items.manual is not None and self._config.items.manual.is_on():
            return "Manual"

        if self._get_on_off_conditions_met():
            return "Auto_On"
        if self._current_above_threshold():
            return "Auto_WaitCurrent"
        return "Auto_Off"

    def _update_openhab_state(self) -> None:
        """Update OpenHAB state item and other states.

        This should method should be set to "after_state_change" of the state machine.
        """
        if self.state != self._previous_state:
            super()._update_openhab_state()
            self._instance_logger.debug(f"State change: {self._previous_state} -> {self.state}")

            self._set_switch_state()
            self._previous_state = self.state

    def on_enter_Auto_Init(self) -> None:  # noqa: N802
        """Callback, which is called on enter of init state."""
        if self._get_on_off_conditions_met():
            self.to_Auto_On()
        else:
            self.to_Auto_Off()

    def _set_switch_state(self) -> None:
        """Set switch state."""
        if self.state == "Auto_On":
            self._switch_observer.send_command("ON")
        elif self.state == "Auto_Off":
            self._switch_observer.send_command("OFF")

    def _current_above_threshold(self) -> bool:
        """Current is above threshold.

        Returns:
            True if current is above threshold
        """
        if self._config.items.current is None or self._config.items.current.value is None:
            return False

        return self._config.items.current.value > self._config.parameter.current_threshold

    def _get_on_off_conditions_met(self) -> bool:
        """Check if on/off conditions are met.

        Returns:
            True if switch should be ON, else False
        """
        external_req = self._config.items.external_request is not None and self._config.items.external_request.is_on()
        present = self._config.items.presence_state is not None and self._config.items.presence_state == PresenceState.PRESENCE.value
        awake = self._config.items.sleeping_state is not None and self._config.items.sleeping_state == SleepState.AWAKE.value

        return external_req or (present and awake)

    def _conditions_changed(self) -> None:
        """Sleep, presence or external state changed."""
        if self._get_on_off_conditions_met():
            self.on_conditions_met()
        else:
            self.off_conditions_met()

    def _cb_max_on_countdown(self) -> None:
        """Callback which is triggered if max on time is reached."""
        if self._max_on_countdown:
            if self._config.items.external_request is not None and self._config.items.external_request.is_on():
                self._switch_off_after_external_req = True
            else:
                self.max_on_countdown()

    def _cb_switch(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if switch changed.

        Args:
          event: event which triggered this callback
        """
        if self._max_on_countdown is not None:
            if event.value == "ON":
                self._max_on_countdown.reset()
            else:
                self._max_on_countdown.stop()

    def _cb_hand(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:  # noqa: ARG002
        """Callback, which is triggered by the state observer if a manual change was detected.

        Args:
          event: event which triggered this callback.
        """
        self.hand_detected()

    def _cb_manual(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is triggered if the manual switch has a state change event.

        Args:
          event: trigger event
        """
        if event.value == "ON":
            self.manual_on()
        else:
            self.manual_off()

    def _cb_presence_state(self, _: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if presence_state changed."""
        self._conditions_changed()

    def _cb_sleeping_state(self, _: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if sleeping state changed."""
        self._conditions_changed()

    def _cb_external_request(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if external request changed."""
        if event.value == "OFF" and self._switch_off_after_external_req:
            self.to_Auto_Off()
        else:
            self._conditions_changed()
        self._switch_off_after_external_req = False

    def _cb_current_changed(self, _: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if the current value changed."""
        if self.state == "Auto_WaitCurrent" and not self._current_above_threshold():
            self.current_below_threshold()

        if self.state == "Auto_WaitCurrentExtended" and self._current_above_threshold():
            self.current_above_threshold()



"""Test heating rules."""

import collections
import datetime
import unittest.mock

import HABApp
import HABApp.rule.scheduler.job_ctrl

import habapp_rules.actors.config.heating
import habapp_rules.actors.heating
import habapp_rules.core.exceptions
import tests.helper.oh_item
import tests.helper.test_case_base


class TestKnxHeating(tests.helper.test_case_base.TestCaseBase):
    """Test KnxHeating."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Temperature_OH", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Temperature_KNX", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Offset", None)

        self._config = habapp_rules.actors.config.heating.KnxHeatingConfig(
            items=habapp_rules.actors.config.heating.KnxHeatingItems(virtual_temperature="Unittest_Temperature_OH", actor_feedback_temperature="Unittest_Temperature_KNX", temperature_offset="Unittest_Offset")
        )

        self._rule = habapp_rules.actors.heating.KnxHeating(self._config)

    def test_init(self) -> None:
        """Test __init__."""
        rule = habapp_rules.actors.heating.KnxHeating(self._config)
        self.assertIsNone(rule._temperature)

        tests.helper.oh_item.set_state("Unittest_Temperature_KNX", 42)
        rule = habapp_rules.actors.heating.KnxHeating(self._config)
        self.assertEqual(42, rule._temperature)

    def test_feedback_temperature_changed(self) -> None:
        """Test _cb_actor_feedback_temperature_changed."""
        tests.helper.oh_item.assert_value("Unittest_Temperature_OH", None)
        tests.helper.oh_item.item_state_change_event("Unittest_Temperature_KNX", 42)
        tests.helper.oh_item.assert_value("Unittest_Temperature_OH", 42)
        self.assertEqual(42, self._rule._temperature)

    def test_virtual_temperature_command(self) -> None:
        """Test _cb_virtual_temperature_command."""
        # _temperature and temperature_offset are None
        self.assertIsNone(self._rule._temperature)
        tests.helper.oh_item.assert_value("Unittest_Offset", None)

        tests.helper.oh_item.item_command_event("Unittest_Temperature_OH", 42)

        self.assertEqual(42, self._rule._temperature)
        tests.helper.oh_item.assert_value("Unittest_Offset", 0)

        TestCase = collections.namedtuple("TestCase", "event_value, rule_temperature, offset_value, expected_new_offset")

        test_cases = [
            TestCase(20, 19, 0, 1),
            TestCase(21.5, 19, 0, 2.5),
            TestCase(19, 20, 0, -1),
            TestCase(15.5, 20, 0, -4.5),
            TestCase(20, 19, 1, 2),
            TestCase(21.5, 19, 1.5, 4),
            TestCase(22.4, 19, 1, 4.4),
            TestCase(20, 19, -1, 0),
            TestCase(21.5, 19, -1.5, 1),
            TestCase(22.4, 19, -1, 2.4),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                self._rule._temperature = test_case.rule_temperature
                tests.helper.oh_item.set_state("Unittest_Offset", test_case.offset_value)

                tests.helper.oh_item.item_command_event("Unittest_Temperature_OH", test_case.event_value)

                self.assertEqual(test_case.expected_new_offset, round(self._rule._config.items.temperature_offset.value, 1))
                self.assertEqual(test_case.event_value, self._rule._temperature)


class TestHeatingActive(tests.helper.test_case_base.TestCaseBase):
    """Test HeatingActive."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_ctr_value_1", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_ctr_value_2", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_heating_active_1", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_heating_active_2", None)

        self._config_1 = habapp_rules.actors.config.heating.HeatingActiveConfig(items=habapp_rules.actors.config.heating.HeatingActiveItems(control_values=["Unittest_ctr_value_1", "Unittest_ctr_value_2"], output="Unittest_heating_active_1"))
        self._config_2 = habapp_rules.actors.config.heating.HeatingActiveConfig(
            items=habapp_rules.actors.config.heating.HeatingActiveItems(control_values=["Unittest_ctr_value_1", "Unittest_ctr_value_2"], output="Unittest_heating_active_2"),
            parameter=habapp_rules.actors.config.heating.HeatingActiveParameter(extended_active_time=datetime.timedelta(seconds=10), threshold=20),
        )

        self._rule_1 = habapp_rules.actors.heating.HeatingActive(self._config_1)
        self._rule_2 = habapp_rules.actors.heating.HeatingActive(self._config_2)

    def test_init(self) -> None:
        """Test __init__."""
        self.assertTrue(isinstance(self._rule_1._extended_lock, HABApp.rule.scheduler.job_ctrl.CountdownJobControl))
        self.assertIsNone(self._rule_1._extended_lock.next_run_datetime)
        tests.helper.oh_item.assert_value("Unittest_heating_active_1", "OFF")
        tests.helper.oh_item.assert_value("Unittest_heating_active_2", "OFF")
        self.assertIn("_cb_lock_end", self._rule_1._extended_lock._job.executor._func.name)

    def test_init_when_output_is_on(self) -> None:
        """Test __init__ when output is on."""
        TestCase = collections.namedtuple("TestCase", "state, reset_called")

        test_cases = [TestCase("ON", True), TestCase("OFF", False), TestCase(None, False)]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                tests.helper.oh_item.set_state("Unittest_heating_active_1", test_case.state)
                with unittest.mock.patch("HABApp.rule.scheduler.job_builder.HABAppJobBuilder.countdown") as run_countdown_mock:
                    habapp_rules.actors.heating.HeatingActive(self._config_1)
                if test_case.reset_called:
                    run_countdown_mock.return_value.reset.assert_called_once()
                else:
                    run_countdown_mock.return_value.reset.assert_not_called()

    def test_set_output(self) -> None:
        """Test _set_output."""
        tests.helper.oh_item.assert_value("Unittest_heating_active_1", "OFF")

        # ctr_value_1 changes to 0
        tests.helper.oh_item.item_state_change_event("Unittest_ctr_value_1", 0)
        tests.helper.oh_item.assert_value("Unittest_heating_active_1", "OFF")
        tests.helper.oh_item.assert_value("Unittest_heating_active_2", "OFF")

        # ctr_value_1 changes to 10
        tests.helper.oh_item.item_state_change_event("Unittest_ctr_value_1", 10)
        tests.helper.oh_item.assert_value("Unittest_heating_active_1", "ON")
        tests.helper.oh_item.assert_value("Unittest_heating_active_2", "OFF")

        # ctr_value_2 changes to 21
        tests.helper.oh_item.item_state_change_event("Unittest_ctr_value_2", 21)
        tests.helper.oh_item.assert_value("Unittest_heating_active_1", "ON")
        tests.helper.oh_item.assert_value("Unittest_heating_active_2", "ON")

        # both change to 0
        tests.helper.oh_item.item_state_change_event("Unittest_ctr_value_1", 0)
        tests.helper.oh_item.item_state_change_event("Unittest_ctr_value_2", 0)
        tests.helper.oh_item.assert_value("Unittest_heating_active_1", "ON")
        tests.helper.oh_item.assert_value("Unittest_heating_active_2", "ON")

        # lock period ends
        self._rule_1._cb_lock_end()
        self._rule_2._cb_lock_end()
        tests.helper.oh_item.assert_value("Unittest_heating_active_1", "OFF")
        tests.helper.oh_item.assert_value("Unittest_heating_active_2", "OFF")



"""Define graph machine classes."""

import os
from functools import partial

import transitions.extensions.states
from transitions.extensions.diagrams_graphviz import Graph, _filter_states

try:
    import graphviz as pgv
except ImportError:
    pgv = None

os.environ["PATH"] += r"C:\Program Files\Graphviz\bin"


class FakeModel:
    """This class is used as fake model for graph creation."""


def get_graph_with_previous_state(self: Graph, title: str | None = None, roi_state: str | None = None) -> object:
    """Monkey patch for transtitions.extentions.diagrams_graphviz.Graph.get_graph, which also adds all previous states.

    Args:
        self: graph object
        title: title of graph
        roi_state: region of interest - state

    Returns:
        graph
    """
    title = title or self.machine.title

    fsm_graph = pgv.Digraph(
        name=title,
        node_attr=self.machine.style_attributes["node"]["default"],
        edge_attr=self.machine.style_attributes["edge"]["default"],
        graph_attr=self.machine.style_attributes["graph"]["default"],
    )
    fsm_graph.graph_attr.update(**self.machine.machine_attributes)
    fsm_graph.graph_attr["label"] = title
    # For each state, draw a circle
    states, trans = self._get_elements()
    if roi_state:
        trans = [t for t in trans if t["source"] == roi_state or t["dest"] == roi_state or self.custom_styles["edge"][t["source"]][t["dest"]]]
        state_names = [t for trans in trans for t in [trans["source"], trans.get("dest", trans["source"])]]
        state_names += [k for k, style in self.custom_styles["node"].items() if style]
        states = _filter_states(states, state_names, self.machine.state_cls)
    self._add_nodes(states, fsm_graph)
    self._add_edges(trans, fsm_graph)
    fsm_graph.draw = partial(self.draw, fsm_graph)
    return fsm_graph


# monkey patching of Graph method
Graph.get_graph = get_graph_with_previous_state


@transitions.extensions.states.add_state_features(transitions.extensions.states.Timeout)
class GraphMachineTimer(transitions.extensions.GraphMachine):
    """GraphMachine with Timer."""


@transitions.extensions.states.add_state_features(transitions.extensions.states.Timeout)
class HierarchicalGraphMachineTimer(transitions.extensions.HierarchicalGraphMachine):
    """HierarchicalGraphMachine with Timer."""



"""Config models for sleep rules."""

import datetime

import HABApp
import pydantic

import habapp_rules.core.pydantic_base


class SleepItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for sleep detection."""

    sleep: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="sleep item")
    sleep_request: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="sleep request item")
    lock: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="lock item")
    lock_request: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="lock request item")
    display_text: HABApp.openhab.items.StringItem | None = pydantic.Field(None, description="display text item")
    state: HABApp.openhab.items.StringItem = pydantic.Field(..., description="state item")


class SleepConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for sleep detection."""

    items: SleepItems = pydantic.Field(..., description="items for sleep state")
    parameter: None = None


# LINK SLEEP ##############################


class LinkSleepItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for sleep detection."""

    sleep_master: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="sleep item of the the master item which will link it's state to the slave items")
    sleep_request_slaves: list[HABApp.openhab.items.SwitchItem] = pydantic.Field(..., description="list of sleep request items of the slaves")
    link_active_feedback: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="item which is ON if link is active or OFF if link is not active anymore")


class LinkSleepParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Config for sleep detection."""

    link_time_start: datetime.time = pydantic.Field(datetime.time(0), description="Start time when the linking is active")
    link_time_end: datetime.time = pydantic.Field(datetime.time(23, 59, 59), description="End time when the linking is not active anymore")


class LinkSleepConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for sleep detection."""

    items: LinkSleepItems = pydantic.Field(..., description="items for sleep state")
    parameter: LinkSleepParameter = pydantic.Field(LinkSleepParameter(), description="parameter for link sleep")



"""Implementations for observing states of switch / dimmer / roller shutter."""

from __future__ import annotations

import abc
import logging
import threading
import time
from collections.abc import Callable

import HABApp
import HABApp.openhab.items

import habapp_rules.core.exceptions
import habapp_rules.core.logger
import habapp_rules.core.timeout_list

LOGGER = logging.getLogger(__name__)

EventTypes = HABApp.openhab.events.ItemStateChangedEvent | HABApp.openhab.events.ItemCommandEvent
CallbackType = Callable[[EventTypes], None]


class _StateObserverBase(HABApp.Rule, abc.ABC):
    """Base class for observer classes."""

    def __init__(self, item_name: str, control_names: list[str] | None = None, group_names: list[str] | None = None, value_tolerance: float = 0) -> None:
        """Init state observer for switch item.

        Args:
            item_name: Name of observed item
            control_names: list of control items.
            group_names: list of group items where the item is a part of. Group item type must match with type of item_name
            value_tolerance: used by all observers which handle numbers. It can be used to allow a difference when comparing new and old values.
        """
        self._value_tolerance = value_tolerance

        HABApp.Rule.__init__(self)
        self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, item_name)

        self._last_manual_event = HABApp.openhab.events.ItemCommandEvent("", None)

        self._item = HABApp.openhab.items.OpenhabItem.get_item(item_name)

        self.__control_items = [HABApp.openhab.items.OpenhabItem.get_item(name) for name in control_names] if control_names else []
        self.__group_items = [HABApp.openhab.items.OpenhabItem.get_item(name) for name in group_names] if group_names else []
        self.__check_item_types()

        self._value = self._item.value
        self._group_last_event = 0

        self._item.listen_event(self._cb_item, HABApp.openhab.events.ItemStateChangedEventFilter())
        for control_item in self.__control_items:
            control_item.listen_event(self._cb_control_item, HABApp.openhab.events.ItemCommandEventFilter())
        for group_item in self.__group_items:
            group_item.listen_event(self._cb_group_item, HABApp.openhab.events.ItemStateUpdatedEventFilter())

    @property
    def value(self) -> float | bool:
        """Get the current state / value of the observed item.

        Returns:
            Current value of the observed item
        """
        return self._value

    @property
    def last_manual_event(self) -> EventTypes:
        """Get the last manual event.

        Returns:
            Last manual event
        """
        return self._last_manual_event

    def __check_item_types(self) -> None:
        """Check if all command and control items have the correct type.

        Raises:
            TypeError: if one item has the wrong type
        """
        target_type = type(self._item)

        wrong_types = [f"{item.name} <{type(item).__name__}>" for item in self.__control_items + self.__group_items if not isinstance(item, target_type)]

        if wrong_types:
            self._instance_logger.error(msg := f"Found items with wrong item type. Expected: {target_type.__name__}. Wrong: {' | '.join(wrong_types)}")
            raise TypeError(msg)

    @abc.abstractmethod
    def send_command(self, value: float | str) -> None:
        """Send brightness command to light (this should be used by rules, to not trigger a manual action).

        Args:
            value: Value to send to the light

        Raises:
            ValueError: if value has wrong format
        """

    def _cb_item(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is called if a value change of the light item was detected.

        Args:
            event: event, which triggered this callback
        """
        self._check_manual(event)

    def _cb_group_item(self, event: HABApp.openhab.events.ItemStateUpdatedEvent) -> None:
        """Callback, which is called if a value change of the light item was detected.

        Args:
            event: event, which triggered this callback
        """
        if event.value in {"ON", "OFF"} and time.time() - self._group_last_event > 0.3:  # this is some kind of workaround. For some reason all events are doubled. # noqa: PLR2004
            self._group_last_event = time.time()
            self._check_manual(event)

    @abc.abstractmethod
    def _cb_control_item(self, event: HABApp.openhab.events.ItemCommandEvent) -> None:
        """Callback, which is called if a command event of one of the control items was detected.

        Args:
            event: event, which triggered this callback
        """

    @abc.abstractmethod
    def _check_manual(self, event: HABApp.openhab.events.ItemStateChangedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:
        """Check if light was triggered by a manual action.

        Args:
            event: event which triggered this method. This will be forwarded to the callback

        Raises:
            ValueError: if event is not supported
        """

    def _trigger_callback(self, cb_name: str, event: EventTypes) -> None:
        """Trigger a manual detected callback.

        Args:
            cb_name: name of callback method
            event: event which triggered the callback
        """
        self._last_manual_event = event
        callback: CallbackType = getattr(self, cb_name)
        if callback is not None:
            callback(event)

    def _values_different_with_tolerance(self, value_1: float, value_2: float) -> bool:
        """Check if values are different, including the difference.

        Args:
            value_1: first value
            value_2: second value

        Returns:
            true if values are different (including the tolerance), false if not
        """
        return abs((value_1 or 0) - (value_2 or 0)) > self._value_tolerance


class StateObserverSwitch(_StateObserverBase):
    """Class to observe the on/off state of a switch item.

    This class is normally not used standalone. Anyway here is an example config:

    # KNX-things:
    Thing device T00_99_OpenHab_DimmerSwitch "KNX OpenHAB switch observer"{
        Type switch             : switch             "Switch"             [ switch="1/1/10" ]
    }

    # Items:
    Switch    I01_01_Switch    "Switch"      {channel="knx:device:bridge:T00_99_OpenHab_DimmerSwitch:switch"}

    # Rule init:
    habapp_rules.actors.state_observer.StateObserverSwitch("I01_01_Switch", callback_on, callback_off)
    """

    def __init__(self, item_name: str, cb_on: CallbackType, cb_off: CallbackType) -> None:
        """Init state observer for switch item.

        Args:
            item_name: Name of switch item
            cb_on: callback which should be called if manual_on was detected
            cb_off: callback which should be called if manual_off was detected
        """
        self._cb_on = cb_on
        self._cb_off = cb_off
        _StateObserverBase.__init__(self, item_name)
        self._value = self._item.value

    def _check_manual(self, event: HABApp.openhab.events.ItemStateChangedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:
        """Check if light was triggered by a manual action.

        Args:
            event: event which triggered this method. This will be forwarded to the callback

        Raises:
            ValueError: if event is not supported
        """
        if event.value == "ON" and not self._value:
            self._value = True
            self._trigger_callback("_cb_on", event)

        elif event.value == "OFF" and self._value:
            self._value = False
            self._trigger_callback("_cb_off", event)

    def _cb_control_item(self, event: HABApp.openhab.events.ItemCommandEvent) -> None:  # not used by StateObserverSwitch
        """Callback, which is called if a command event of one of the control items was detected.

        Args:
            event: event, which triggered this callback
        """

    def send_command(self, value: str) -> None:
        """Send brightness command to light (this should be used by rules, to not trigger a manual action).

        Args:
            value: Value to send to the light

        Raises:
            ValueError: if value has wrong format
        """
        if value == "ON":
            self._value = True

        elif value == "OFF":
            self._value = False
        else:
            msg = f"The given value is not supported for StateObserverSwitch: {value}"
            raise ValueError(msg)

        self._item.oh_send_command(value)


class StateObserverDimmer(_StateObserverBase):
    """Class to observe the on / off / change events of a dimmer item.

    Known limitation: if the items of group_names are KNX-items, the channel types must be dimmer (not dimmer-control)
    This class is normally not used standalone. Anyway here is an example config:

    # KNX-things:
    Thing device T00_99_OpenHab_DimmerObserver "KNX OpenHAB dimmer observer"{
        Type dimmer             : light             "Light"             [ switch="1/1/10", position="1/1/13+<1/1/15" ]
        Type dimmer-control     : light_ctr         "Light control"     [ increaseDecrease="1/1/12"]
        Type dimmer             : light_group       "Light Group"       [ switch="1/1/240", position="1/1/243"]
    }

    # Items:
    Dimmer    I01_01_Light              "Light"             {channel="knx:device:bridge:T00_99_OpenHab_DimmerObserver:light"}
    Dimmer    I01_01_Light_ctr          "Light ctr"         {channel="knx:device:bridge:T00_99_OpenHab_DimmerObserver:light_ctr"}
    Dimmer    I01_01_Light_group        "Light Group"       {channel="knx:device:bridge:T00_99_OpenHab_DimmerObserver:light_group"}

    # Rule init:
    habapp_rules.actors.state_observer.StateObserverDimmer(
                    "I01_01_Light",
                    control_names=["I01_01_Light_ctr"],
                    group_names=["I01_01_Light_group"],
                    cb_on=callback_on,
                    cb_off=callback_off,
                    cb_brightness_change=callback_change)
    """

    def __init__(self, item_name: str, cb_on: CallbackType, cb_off: CallbackType, cb_brightness_change: CallbackType | None = None, control_names: list[str] | None = None, group_names: list[str] | None = None, value_tolerance: float = 0) -> None:
        """Init state observer for dimmer item.

        Args:
            item_name: Name of dimmer item
            cb_on: callback which is called if manual_on was detected
            cb_off: callback which is called if manual_off was detected
            cb_brightness_change: callback which is called if dimmer is on and brightness changed
            control_names: list of control items. They are used to also respond to switch on/off via INCREASE/DECREASE
            group_names: list of group items where the item is a part of. Group item type must match with type of item_name
            value_tolerance: the tolerance can be used to allow a difference when comparing new and old values.
        """
        _StateObserverBase.__init__(self, item_name, control_names, group_names, value_tolerance)

        self._cb_on = cb_on
        self._cb_off = cb_off
        self._cb_brightness_change = cb_brightness_change

    def _check_manual(self, event: HABApp.openhab.events.ItemStateChangedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:
        """Check if light was triggered by a manual action.

        Args:
            event: event which triggered this method. This will be forwarded to the callback
        """
        if isinstance(event.value, int | float):
            if event.value > 0 and (self._value is None or self._value == 0):
                self._value = event.value
                self._trigger_callback("_cb_on", event)

            elif event.value == 0 and (self._value is None or self._value > 0):
                self._value = 0
                self._trigger_callback("_cb_off", event)

            elif self._values_different_with_tolerance(event.value, self._value):
                self._value = event.value
                self._trigger_callback("_cb_brightness_change", event)

        elif event.value == "ON" and (self._value is None or self._value == 0):
            self._value = 100
            self._trigger_callback("_cb_on", event)

        elif event.value == "OFF" and (self._value is None or self._value > 0):
            self._value = 0
            self._trigger_callback("_cb_off", event)

    def _cb_control_item(self, event: HABApp.openhab.events.ItemCommandEvent) -> None:
        """Callback, which is called if a command event of one of the control items was detected.

        Args:
            event: event, which triggered this callback
        """
        if event.value == "INCREASE" and (self._value is None or self._value == 0):
            self._value = 100
            self._trigger_callback("_cb_on", event)

    def send_command(self, value: float | str) -> None:
        """Send brightness command to light (this should be used by rules, to not trigger a manual action).

        Args:
            value: Value to send to the light

        Raises:
            ValueError: if value has wrong format
        """
        if isinstance(value, int | float):
            self._value = value

        elif value == "ON":
            self._value = 100

        elif value == "OFF":
            self._value = 0

        else:
            msg = f"The given value is not supported for StateObserverDimmer: {value}"
            raise ValueError(msg)

        self._item.oh_send_command(value)


class StateObserverRollerShutter(_StateObserverBase):
    """Class to observe manual controls of a roller shutter item.

    This class is normally not used standalone. Anyway, here is an example config:

    # KNX-things:
    Thing device T00_99_OpenHab_RollershutterObserver "KNX OpenHAB rollershutter observer"{
       Type rollershutter             : shading             "Shading"             [ upDown="1/1/10", position="1/1/13+<1/1/15" ]
       Type rollershutter-control     : shading_ctr         "Shading control"     [ upDown="1/1/10", position="1/1/13+<1/1/15" ]
       Type rollershutter             : shading_group       "Shading Group"       [ upDown="1/1/110", position="1/1/113+<1/1/115" ]
    }

    # Items:
    Rollershutter    I_Rollershutter              "Rollershutter"             {channel="knx:device:bridge:T00_99_OpenHab_RollershutterObserver:Rollershutter"}
    Rollershutter    I_Rollershutter_ctr          "Rollershutter ctr"         {channel="knx:device:bridge:T00_99_OpenHab_RollershutterObserver:Rollershutter_ctr"}
    Rollershutter    I_Rollershutter_group        "Rollershutter Group"       {channel="knx:device:bridge:T00_99_OpenHab_RollershutterObserver:Rollershutter_group"}

    # Rule init:
    habapp_rules.actors.state_observer.StateObserverRollerShutter(
                "I_Rollershutter",
                control_names=["I_Rollershutter_ctr"],
                group_names=["I_Rollershutter_group"],
                cb_manual=callback_on
                )
    """

    def __init__(self, item_name: str, cb_manual: CallbackType, control_names: list[str] | None = None, group_names: list[str] | None = None, value_tolerance: float = 0) -> None:
        """Init state observer for dimmer item.

        Args:
            item_name: Name of dimmer item
            cb_manual: callback which is called if a manual interaction was detected
            control_names: list of control items. They are used to also respond to switch on/off via INCREASE/DECREASE
            group_names: list of group items where the item is a part of. Group item type must match with type of item_name
            value_tolerance: the tolerance can be used to allow a difference when comparing new and old values.
        """
        self._value_tolerance = value_tolerance
        _StateObserverBase.__init__(self, item_name, control_names, group_names, value_tolerance)

        self._cb_manual = cb_manual

    def _check_manual(self, event: HABApp.openhab.events.ItemStateChangedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:
        """Check if light was triggered by a manual action.

        Args:
            event: event which triggered this method. This will be forwarded to the callback

        Raises:
            ValueError: if event is not supported
        """
        if isinstance(event.value, int | float) and self._values_different_with_tolerance(event.value, self._value):
            self._value = event.value
            self._trigger_callback("_cb_manual", event)

    def _cb_control_item(self, event: HABApp.openhab.events.ItemCommandEvent) -> None:
        """Callback, which is called if a command event of one of the control items was detected.

        Args:
            event: event, which triggered this callback
        """
        if event.value == "DOWN":
            self._value = 100
            self._trigger_callback("_cb_manual", event)

        elif event.value == "UP":
            self._value = 0
            self._trigger_callback("_cb_manual", event)

    def send_command(self, value: float) -> None:
        """Send brightness command to light (this should be used by rules, to not trigger a manual action).

        Args:
            value: Value to send to the light

        Raises:
            TypeError: if value has wrong format
        """
        if not isinstance(value, int | float):
            msg = f"The given value is not supported for StateObserverDimmer: {value}"
            raise TypeError(msg)

        self._value = value
        self._item.oh_send_command(value)


class StateObserverNumber(_StateObserverBase):
    """Class to observe the state of a number item.

    This class is normally not used standalone. Anyway here is an example config:

    # KNX-things:
    Thing device T00_99_OpenHab_DimmerNumber "KNX OpenHAB number observer"{
        Type number             : number             "Switch"             [ ga="1/1/10" ]
    }

    # Items:
    Number    I01_01_Number    "Switch"      {channel="knx:device:bridge:T00_99_OpenHab_DimmerNumber:number"}

    # Rule init:
    habapp_rules.actors.state_observer.StateObserverNumber("I01_01_Number", callback_value_changed)
    """

    def __init__(self, item_name: str, cb_manual: CallbackType, value_tolerance: float = 0) -> None:
        """Init state observer for switch item.

        Args:
            item_name: Name of switch item
            cb_manual: callback which should be called if manual change was detected
            value_tolerance: the tolerance can be used to allow a difference when comparing new and old values.
        """
        self._cb_manual = cb_manual
        _StateObserverBase.__init__(self, item_name, value_tolerance=value_tolerance)

    def _check_manual(self, event: HABApp.openhab.events.ItemStateChangedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:
        """Check if light was triggered by a manual action.

        Args:
            event: event which triggered this method. This will be forwarded to the callback

        Raises:
            ValueError: if event is not supported
        """
        if self._value is None:
            self._value = event.value
            return

        if self._values_different_with_tolerance(event.value, self._value):
            self._value = event.value
            self._trigger_callback("_cb_manual", event)

    def _cb_control_item(self, event: HABApp.openhab.events.ItemCommandEvent) -> None:  # not used by StateObserverNumber
        """Callback, which is called if a command event of one of the control items was detected.

        Args:
            event: event, which triggered this callback
        """

    def send_command(self, value: float) -> None:
        """Send brightness command to light (this should be used by rules, to not trigger a manual action).

        Args:
            value: Value to send to the light

        Raises:
            TypeError: if value has wrong format
        """
        if not isinstance(value, int | float):
            msg = f"The given value is not supported for StateObserverNumber: {value}"
            raise TypeError(msg)
        self._value = value
        self._item.oh_send_command(value)


class StateObserverSlat(StateObserverNumber):
    """This is only used for the slat value of shading!"""

    def __init__(self, item_name: str, cb_manual: CallbackType, value_tolerance: float = 0) -> None:
        """Init state observer for switch item.

        Args:
            item_name: Name of switch item
            cb_manual: callback which should be called if manual change was detected
            value_tolerance: the tolerance can be used to allow a difference when comparing new and old values.
        """
        self.__timer_manual: threading.Timer | None = None
        StateObserverNumber.__init__(self, item_name, cb_manual, value_tolerance)

    def _check_manual(self, event: HABApp.openhab.events.ItemStateChangedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:
        """Check if light was triggered by a manual action.

        Args:
            event: event which triggered this method. This will be forwarded to the callback

        Raises:
            ValueError: if event is not supported
        """
        self._stop_timer_manual()
        if event.value in {0, 100}:
            self.__timer_manual = threading.Timer(3, self.__cb_check_manual_delayed, [event])
            self.__timer_manual.start()
        else:
            StateObserverNumber._check_manual(self, event)  # noqa: SLF001

    def __cb_check_manual_delayed(self, event: HABApp.openhab.events.ItemStateChangedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:
        """Trigger delayed manual check.

        Args:
            event: event which should be checked
        """
        StateObserverNumber._check_manual(self, event)  # noqa: SLF001

    def _stop_timer_manual(self) -> None:
        """Stop timer if running."""
        if self.__timer_manual:
            self.__timer_manual.cancel()
            self.__timer_manual = None

    def on_rule_removed(self) -> None:
        """Stop timer if rule is removed."""
        self._stop_timer_manual()



import datetime
import logging

import HABApp

LOGGER = logging.getLogger(__name__)


def get_historic_value(item: HABApp.openhab.items.NumberItem, start_time: datetime.datetime) -> float:
    """Get historic value of given Number item.

    Args:
        item: item instance
        start_time: start time to search for the interested value

    Returns:
        historic value of the item
    """
    historic = item.get_persistence_data(start_time=start_time, end_time=start_time + datetime.timedelta(hours=1)).data
    if not historic:
        LOGGER.warning(f"Could not get value of item '{item.name}' of time = {start_time}")
        return 0

    return next(iter(historic.values()))



import json
import logging
import pathlib
import sys

import requests

import habapp_rules

POSITIVE_RESPONSE_CODE = 200
VERSION_LENGTH = 3

logging.basicConfig(
    level=logging.INFO,  # Set the logging level (INFO, DEBUG, etc.)
    format="%(levelname)s: %(message)s",  # Log message format
    handlers=[logging.StreamHandler(sys.stdout)],  # Use sys.stdout for logs
)

LOGGER = logging.getLogger(__name__)


class VersionCheck:
    """Checker to asure that version was updated."""

    def __init__(self, current_version: str, pypi_pkg_name: str, changelog_path: str | None = None) -> None:
        """Create checker.

        Args:
            current_version: Version of the current branch. E.g 1.1.0
            pypi_pkg_name: Name of the PyPi package
            changelog_path: path to changelog file
        """
        self.current_version = current_version
        self.pypi_pkg_name = pypi_pkg_name
        self.changelog_path = pathlib.Path(changelog_path) if changelog_path else None

    def __get_pypi_version(self) -> str:
        """Get the newest version from PyPi.

        Returns:
                Returns the newest version which is released on PyPi

        Raises:
        ConnectionError: if no response was received
        """
        result = requests.get(f"https://pypi.org/pypi/{self.pypi_pkg_name}/json", timeout=10)
        if result.status_code != POSITIVE_RESPONSE_CODE:
            msg = "no response!"
            raise ConnectionError(msg)

        return json.loads(result.text)["info"]["version"]

    @staticmethod
    def __str_to_version(version: str) -> int:
        """Get a integer representation for a given version as string.

        Args:
            version: Version as string. E.g. 1.1.0

        Returns:
            A integer representation of the given version

        Raises:
        AttributeError: if the format of the given version is not correct
        """
        version_parts = version.split(".")

        if len(version_parts) != VERSION_LENGTH or not all(value.isdigit() for value in version_parts):
            msg = f"The format of the given version ({version}) is not correct. Version must have the following format X.X.X"
            raise AttributeError(msg)

        return int(version_parts[0]) * 1000000 + int(version_parts[1]) * 1000 + int(version_parts[2])

    def check_version(self) -> int:
        """Check if version of the current branch is higher than the newest PyPi version.

        Returns:
            0 if branch version is higher than PyPi version. If not -1
        """
        passed = True

        pypi_version = self.__get_pypi_version()
        branch_version = self.current_version

        LOGGER.info(f"PyPi version: {pypi_version} | branch version: {branch_version}")

        if not self.__str_to_version(branch_version) > self.__str_to_version(pypi_version):
            passed = False
            LOGGER.error("Increase version of branch!")

        if self.changelog_path:
            with self.changelog_path.open("r") as changelog_file:
                first_line = changelog_file.readline()

            if branch_version not in first_line:
                LOGGER.error(f"Current version '{branch_version}' must be mentioned in the first line of changelog.md!")
                passed = False

        return 0 if passed else -1


if __name__ == "__main__":
    sys.exit(VersionCheck(habapp_rules.__version__, "habapp_rules", "changelog.md").check_version())



"""Config models for sun rules."""

import logging

import HABApp
import pydantic
import typing_extensions

import habapp_rules.core.pydantic_base

LOGGER = logging.getLogger(__name__)


class _ItemsBase(habapp_rules.core.pydantic_base.ItemBase):
    """Base class for items for sun sensor."""

    output: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="output item")
    threshold: HABApp.openhab.items.NumberItem | None = pydantic.Field(None, description="threshold item")


class BrightnessItems(_ItemsBase):
    """Items for sun sensor which uses brightness items as input."""

    brightness: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="brightness item")


class TemperatureDifferenceItems(_ItemsBase):
    """Items for sun sensor which uses temperature items as input."""

    temperatures: list[HABApp.openhab.items.NumberItem] = pydantic.Field(..., description="temperature items")

    @pydantic.model_validator(mode="after")
    def validate_temperature_items(self) -> typing_extensions.Self:
        """Validate that at least two temperature items are given.

        Returns:
            validated model

        Raises:
            ValueError: if less than two temperature items are given
        """
        if len(self.temperatures) < 2:  # noqa: PLR2004
            msg = "At least two temperature items are required!"
            raise ValueError(msg)
        return self


class BrightnessParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Parameter for sun sensor which uses brightness items as input."""

    threshold: float | None = pydantic.Field(None, description="threshold value")
    hysteresis: float = pydantic.Field(0.0, description="hysteresis value")
    filter_tau: int = pydantic.Field(30 * 60, description="filter constant for the exponential filter. Default is set to 30 minutes")
    filter_instant_increase: bool = pydantic.Field(default=True, description="if set to True, increase of input values will not be filtered")
    filter_instant_decrease: bool = pydantic.Field(default=False, description="if set to True, decrease of input values will not be filtered")
    filtered_signal_groups: list[str] = pydantic.Field([], description="group names where the filtered signal will be added")


class TemperatureDifferenceParameter(BrightnessParameter):
    """Parameter for sun sensor which uses temperature items as input."""

    ignore_old_values_time: int | None = pydantic.Field(None, description="ignores values which are older than the given time in seconds. If None, all values will be taken")


class _ConfigBase(habapp_rules.core.pydantic_base.ConfigBase):
    """Base config model for sun sensor."""

    items: BrightnessItems | TemperatureDifferenceItems = pydantic.Field(..., description="items for sun sensor")
    parameter: BrightnessParameter | TemperatureDifferenceParameter = pydantic.Field(..., description="parameter for sun sensor")

    @pydantic.model_validator(mode="after")
    def validate_threshold(self) -> typing_extensions.Self:
        """Validate threshold.

        Returns:
            validated model

        Raises:
            ValueError: if threshold and parameter are not set
        """
        if (self.items.threshold is None) == (self.parameter.threshold is None):
            msg = "The threshold must be set ether with the parameter or with the item, both are not allowed"
            raise ValueError(msg)
        return self

    @property
    def threshold(self) -> float:
        """Get threshold."""
        if self.parameter.threshold:
            return self.parameter.threshold

        if self.items.threshold.value is None:
            LOGGER.warning("Threshold item has no value set. Setting threshold to infinity")
            return float("inf")

        return self.items.threshold.value


class BrightnessConfig(_ConfigBase):
    """Config model for sun sensor which uses brightness as input."""

    items: BrightnessItems = pydantic.Field(..., description="items for sun sensor which uses brightness as input")
    parameter: BrightnessParameter = pydantic.Field(BrightnessParameter(), description="parameter for sun sensor which uses brightness as input")


class TemperatureDifferenceConfig(_ConfigBase):
    """Config model for sun sensor which uses temperature items as input."""

    items: TemperatureDifferenceItems = pydantic.Field(..., description="items for sun sensor which uses temperature items as input")
    parameter: TemperatureDifferenceParameter = pydantic.Field(TemperatureDifferenceParameter(), description="parameter for sun sensor which uses temperature items as input")


# SunPositionFilter ###############################
class SunPositionWindow(pydantic.BaseModel):
    """Class for defining min / max values for azimuth and elevation."""

    azimuth_min: float = pydantic.Field(..., description="Starting value for azimuth", ge=0.0, le=360.0)
    azimuth_max: float = pydantic.Field(..., description="End value for azimuth", ge=0.0, le=360.0)
    elevation_min: float = pydantic.Field(0.0, description="Starting value for elevation", ge=-90.0, le=90.0)
    elevation_max: float = pydantic.Field(90.0, description="End value for elevation", ge=-90.0, le=90.0)

    def __init__(self, azimuth_min: float, azimuth_max: float, elevation_min: float = 0.0, elevation_max: float = 90.0) -> None:
        """Init of class for defining min / max values for azimuth and elevation.

        Args:
            azimuth_min: minimum azimuth value
            azimuth_max: maximum azimuth value
            elevation_min: minimum elevation value
            elevation_max: maximum elevation value
        """
        super().__init__(azimuth_min=azimuth_min, azimuth_max=azimuth_max, elevation_min=elevation_min, elevation_max=elevation_max)

    @pydantic.model_validator(mode="after")
    def validate_model(self) -> typing_extensions.Self:
        """Validate values.

        Returns:
            validated model
        """
        if self.azimuth_min > self.azimuth_max:
            LOGGER.warning(f"azimuth_min should be smaller than azimuth_max -> min / max will be swapped. Given values: azimuth_min = {self.azimuth_min} | azimuth_max = {self.azimuth_max}")
            min_orig = self.azimuth_min
            max_orig = self.azimuth_max
            self.azimuth_min = max_orig
            self.azimuth_max = min_orig

        if self.elevation_min > self.elevation_max:
            LOGGER.warning(f"elevation_min should be smaller than elevation_max -> min / max will be swapped. Given values: elevation_min = {self.elevation_min} | elevation_max = {self.elevation_max}")
            min_orig = self.elevation_min
            max_orig = self.elevation_max
            self.elevation_min = max_orig
            self.elevation_max = min_orig
        return self


class SunPositionItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for sun position filter."""

    azimuth: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="sun azimuth item")
    elevation: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="sun elevation item")
    input: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="input item (sun protection required)")
    output: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="output item (sun protection required and sun in the configured azimuth / elevation window)")


class SunPositionParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Parameter for sun position filter."""

    sun_position_window: SunPositionWindow | list[SunPositionWindow] = pydantic.Field(..., description="sun position window, where the sun hits the target")

    @property
    def sun_position_windows(self) -> list[SunPositionWindow]:
        """Get sun position windows."""
        return self.sun_position_window if isinstance(self.sun_position_window, list) else [self.sun_position_window]


class SunPositionConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config model for sun position filter."""

    items: SunPositionItems = pydantic.Field(..., description="items for sun position filter")
    parameter: SunPositionParameter = pydantic.Field(..., description="parameter for sun position filter")


class WinterFilterItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for WinterFilter."""

    sun: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="sun is shining")
    output: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="output item")
    heating_active: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="heating is active")
    presence_state: HABApp.openhab.items.StringItem | None = pydantic.Field(None, description="presence state")


class WinterFilterConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config model for WinterFilter."""

    items: WinterFilterItems = pydantic.Field(..., description="items for sun position filter")
    parameter: None = pydantic.Field(None, description="parameter for sun position filter")



"""Config models for astro rules."""

import HABApp.openhab.items
import pydantic

import habapp_rules.core.pydantic_base


class _NightDayItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for night day."""

    elevation: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="Elevation of the sun")


class SetDayItems(_NightDayItems):
    """Items for setting day."""

    day: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="Item which should be set to ON after dawn and OFF after dusk")


class SetNightItems(_NightDayItems):
    """Items for setting night."""

    night: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="Item which should be set to ON after dusk and OFF after dawn")


class SetDayParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Parameter for setting day."""

    elevation_threshold: float = pydantic.Field(0.0, description="Threshold value for elevation. If the sun elevation is greater than the threshold, the day item will be set to ON")


class SetNightParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Parameter for setting night."""

    elevation_threshold: float = pydantic.Field(-8.0, description="Threshold value for elevation. If the sun elevation is greater than the threshold, the night item will be set to ON")


class SetDayConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for setting day."""

    items: SetDayItems = pydantic.Field(..., description="Items for setting day")
    parameter: SetDayParameter = pydantic.Field(SetDayParameter(), description="Parameter for setting day")


class SetNightConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for setting night."""

    items: SetNightItems = pydantic.Field(..., description="Items for setting night")
    parameter: SetNightParameter = pydantic.Field(SetNightParameter(), description="Parameter for setting night")
